#ifndef MESH_H_
#define MESH_H_

#include <Eigen/Core>
#include <Eigen/Dense>
#include <vector>

class Vertex;
class Edge;
class Face;

class Vertex {
public:
    Edge *edge;        // an outgoing edge of this vertex
    unsigned int id;   // id# of this vertex (unique for this Cell)
};

class Edge {
public:
    unsigned int index; // The index of this edge in its quad-edge structure.
    Edge *next;     //The next ccw edge around (from) the origin of this edge.
    unsigned int id;    // id# of the edge.
    Vertex *vertex;     // The origin vertex of this edge, if prime.
    Face *face;         // The target face of this edge, if dual.
};

class Face {
public:
    Edge *edge;        // a ccw-oriented edge of this face
    unsigned int id;   // id# of this face (unique for this Cell)
};

class Mesh {
public:
	Mesh();
	// Construct the quad-edge structure.
	// The input contains
	//    Mesh-> V - An array of vertices.
	//    Mesh-> F - An array of faces.
	// The output contains
	//    Mesh-> vertices, edges, faces - quad-edge structure
	void BuildQuadEdges();

	// Compute the vertex normals for the mesh.
	// The input contains
	//    Mesh-> V - An array of vertices.
	//    Mesh-> F - An array of faces.
	// The output contains
	//    Mesh-> N - An array of vertex normals.
	void ComputeVertexNormals();

	// Label the vertex colors for the mesh based on regions.
	// The input contains
	//    Mesh-> V - An array of vertices.
	//    vertex_groups - An array of regions, each contains
	//                    an array of vertex indices.
	// The output contains
	//    C - An array of vertex colors.
	void ComputeVertexColors(
		const std::vector<std::vector<int> >& vertex_groups,
		Eigen::MatrixXd& C);

	// Quad edge structure of the mesh
	std::vector<Vertex> vertices;
	std::vector<Edge> edges;
	std::vector<Face> faces;

	// Vertices, Faces and Normals array of the mesh
	Eigen::MatrixXd V;
	Eigen::MatrixXi F;
	Eigen::MatrixXd N;
};

#endif