#ifndef PRIMITIVE_H_
#define PRIMITIVE_H_

#include <Eigen/Core>
#include <set>
#include <vector>

// Region signature structure that contains
//    covariance - Matrix C of the region.
//    null_space - Null space of matrix C.
//    num_points - Number of points in the region.
struct Signature
{
	Eigen::MatrixXd covariance;
	Eigen::MatrixXd null_space;
	int num_points;
};

// Compute the signature of the region.
// The input contains
//     V - An array of vertices.
//     N - An array of vertex normals.
//     vertex_indices - An array of vertex indices
//                      belonging to the region.
// The output contains
//     signature - Region signature.
void EstimatePrimitive(
	const Eigen::MatrixXd& V,
	const Eigen::MatrixXd& N,
	const std::set<int>& vertex_indices,
	Signature& signature);

// Return false if the regions of s1 and s2 should not
// be merged. Otherwise update the signature of s2
// as the merged region of s1 and s2, and return true.
// The input contains
//     s1 - Region signature of the first region.
//     s2 - Region signature of the second region.
//     expected_dimension - The expected dimension of
//                          the merged shape.
// The output contains
//     s2 - Region signature of the merged region.
// Return
//     True/False - whether the regions are merged.
bool MergePrimitive(
	const Signature& s1,
	Signature& s2,
	int expected_dimension);

// Return the condition number for merging two null
// spaces A and B with the given shape dimension.
// The input contains
//     A - Null space of the first region.
//     B - Null space of the second region.
//     expected_dimension - The expected dimension of
//                          the merged shape (k).
// Return
//     The condition number for the (k+1)-th dimension.
double SlippageCondition(
	const Eigen::MatrixXd& A,
	const Eigen::MatrixXd& B,
	int expected_dimension);

#endif