import os, sys
BASE_DIR = os.path.dirname(__file__)
sys.path.append(BASE_DIR)
sys.path.append(os.path.join(BASE_DIR, '..'))
from primitive import primitives
from primitive.util import *

import tensorflow as tf
import numpy as np
from scipy.optimize import linear_sum_assignment

DIVISION_EPS = 1e-10

def nn_filter_W(W, W_null_threshold=0.005):
    # for SPFN & DPPN output, we do two postprocessings: 1) nullify columns with too few points 2) make entries of W binary

    n_points = tf.shape(W)[1]
    n_max_instances = W.get_shape()[2] # n_max_instances should not be dynamic

    W_column_sum = tf.reduce_sum(W, axis=1) # BxK
    null_mask = W_column_sum < tf.to_float(n_points) * W_null_threshold # BxK
    null_mask_W_like = tf.tile(tf.expand_dims(null_mask, axis=1), [1, n_points, 1]) # BxNxK
    W = tf.where(null_mask_W_like, tf.zeros_like(W), tf.one_hot(tf.argmax(W, axis=2), depth=n_max_instances, dtype=tf.float32))

    return W

def evaluate(pred_dict, gt_dict, is_eval, P_in=None):
    '''
        Input: 
            pred_dict should contain:
                - W: BxNxK, segmentation instances. Allow zero rows to indicate unassigned points.
                - normal_per_point: BxNx3, normal per point
                - type_per_point: type per points
                    - This should be logit of shape BxNxT if is_eval=False, and actual value of shape BxN otherwise
                    - can contain -1
                - parameters - a dict, each entry is a BxKx... tensor
            gt_dict should be obtained from calling create_gt_dict
            P_in - BxNx3 is the input point cloud, used only when is_eval=True

        Returns: {loss_dict, matching_indices} + stats from calculate_eval_stats(), where
            - loss_dict contains:
                - normal_loss: B, averaged over all N points
                - type_loss: B, averaged over all N points. 
                    - This is cross entropy loss during training, and accuracy during test time
                - miou_loss: BxK, mean IoU loss for each matched instances
                - residue_loss: BxK, residue loss for each instance
                - parameter_loss: BxK, parameter loss for each instance
                - avg_miou_loss: B
                - avg_residue_loss: B
                - avg_parameter_loss: B
            - matching_indices: BxK, where (b,k)th ground truth primitive is matched with (b, matching_indices[b, k])
    '''
    # dimension tensors
    W = pred_dict['W']
    batch_size = tf.shape(W)[0] 
    n_points = tf.shape(W)[1]
    n_max_instances = W.get_shape()[2] # n_max_instances should not be dynamic
    n_registered_primitives = primitives.get_n_registered_primitives()

    if is_eval:
        # at evaluation, want W to be binary and filtered (if is from nn)
        W = nn_filter_W(W)

    # shortcuts
    # note that I_gt can contain -1, indicating instance of unknown primitive type
    I_gt = gt_dict['instance_per_point'] # BxN
    T_gt = gt_dict['type_per_instance'] # BxK

    n_instances_gt = tf.reduce_max(I_gt, axis=1) + 1 # only count known primitive type instances, as -1 will be ignored
    mask_gt = tf.sequence_mask(n_instances_gt, maxlen=n_max_instances) # BxK, mask_gt[b, k] = 1 iff instace k is present in the ground truth batch b

    ############################
    # TODO Task 2.4: Implement matching and losses.
    matching_indices = None
    miou_loss = tf.zeros(shape=[], dtype=tf.float32)
    normal_loss = tf.zeros(shape=[], dtype=tf.float32)
    per_point_type_loss = tf.zeros(shape=[], dtype=tf.float32)

    residue_losses = [] # a length T array of BxK tensors
    parameter_losses = [] # a length T array of BxK tensors

    for primitive in primitives.get_all_primitives_classes():
        residue_per_point_loss = primitive.compute_residue_loss(pred_dict['parameters'], gt_dict['points_per_instance'], matching_indices) # BxKxN'
        residue_losses.append(tf.reduce_mean(residue_per_point_loss, axis=2))
        parameter_loss = primitive.compute_parameter_loss(pred_dict['parameters'], gt_dict['parameters'], matching_indices, angle_diff=is_eval)
        if parameter_loss is None:
            parameter_loss = tf.zeros(dtype=tf.float32, shape=[batch_size, n_max_instances])
        parameter_losses.append(parameter_loss)
    ############################
    
    residue_losses = tf.stack(residue_losses, axis=2)
    parameter_losses = tf.stack(parameter_losses, axis=2)

    # Aggregate losses across fitters
    residue_loss = aggregate_loss_from_stacked(residue_losses, T_gt) # BxK
    parameter_loss = aggregate_loss_from_stacked(parameter_losses, T_gt) # BxK

    loss_dict = {
        'normal_loss': normal_loss,
        'type_loss': per_point_type_loss,
        'miou_loss': miou_loss,
        'residue_loss': residue_loss,
        'parameter_loss': parameter_loss,
        'avg_miou_loss': reduce_mean_masked_instance(miou_loss, mask_gt),
        'avg_residue_loss': reduce_mean_masked_instance(residue_loss, mask_gt),
        'avg_parameter_loss': reduce_mean_masked_instance(parameter_loss, mask_gt),
    }

    result = {'loss_dict': loss_dict, 'matching_indices': matching_indices}

    return result



def aggregate_loss_from_stacked(loss_stacked, T_gt):
    # loss_stacked - BxKxT, T_gt - BxK
    # out[b, k] = loss_stacked[b, k, T_gt[b, k]]
    B = tf.shape(loss_stacked)[0]
    K = tf.shape(loss_stacked)[1]
    indices_0 = tf.tile(tf.expand_dims(tf.range(B), axis=1), multiples=[1, K]) # BxK
    indices_1 = tf.tile(tf.expand_dims(tf.range(K), axis=0), multiples=[B, 1]) # BxK
    indices = tf.stack([indices_0, indices_1, T_gt], axis=2) # BxKx3
    return tf.gather_nd(loss_stacked, indices=indices)


def reduce_mean_masked_instance(loss, mask_gt):
    # loss: BxK
    loss = tf.where(mask_gt, loss, tf.zeros_like(loss))
    reduced_loss = tf.reduce_sum(loss, axis=1) # B
    denom = tf.reduce_sum(tf.to_float(mask_gt), axis=1) # B
    return tf.where(denom > 0, reduced_loss / denom, tf.zeros_like(reduced_loss)) # B

def hungarian_matching(W_pred, I_gt):
    # This non-tf function does not backprob gradient, only output matching indices
    # W_pred - BxNxK
    # I_gt - BxN, may contain -1's
    # Output: matching_indices - BxK, where (b,k)th ground truth primitive is matched with (b, matching_indices[b, k])
    #   where only n_gt_labels entries on each row have meaning. The matching does not include gt background instance
    batch_size = I_gt.shape[0]
    n_points = I_gt.shape[1]
    n_max_labels = W_pred.shape[2]

    matching_indices = np.zeros([batch_size, n_max_labels], dtype=np.int32)
    for b in range(batch_size):
        # assuming I_gt[b] does not have gap
        n_gt_labels = np.max(I_gt[b]) + 1 # this is K'
        W_gt = np.zeros([n_points, n_gt_labels + 1]) # HACK: add an extra column to contain -1's
        W_gt[np.arange(n_points), I_gt[b]] = 1.0 # NxK'
        
        dot = np.sum(np.expand_dims(W_gt, axis=2) * np.expand_dims(W_pred[b], axis=1), axis=0) # K'xK
        denominator = np.expand_dims(np.sum(W_gt, axis=0), axis=1) + np.expand_dims(np.sum(W_pred[b], axis=0), axis=0) - dot
        cost = dot / np.maximum(denominator, DIVISION_EPS) # K'xK
        cost = cost[:n_gt_labels, :] # remove last row, corresponding to matching gt background instance

        _, col_ind = linear_sum_assignment(-cost) # want max solution
        matching_indices[b, :n_gt_labels] = col_ind

    return matching_indices

