import os, sys

from primitive import primitives
import tensorflow as tf
import evaluation
import architecture
import time

import numpy as np
import h5py
import pickle

class Network(object):
    def __init__(self, n_max_instances):
        self.n_max_instances = n_max_instances

        self.graph = tf.Graph()
        with self.graph.as_default():
            self.global_step = tf.Variable(0)

            self.is_training = tf.placeholder(dtype=tf.bool, shape=[])
            ############################
            # TODO Task 2.1: Create placeholders followed by instruction.
            self.P = None
            self.gt_dict = {}
            self.gt_dict['parameters'] = {}
            for primitive in primitives.get_all_primitives_classes():
                primitive.insert_gt_placeholders(self.gt_dict['parameters'], n_max_instances=n_max_instances)
            ############################

            self.batch_size = tf.shape(self.P)[0]

            self.pred_dict = architecture.get_per_point_model(
                scope='SPFN', 
                P=self.P, 
                n_max_instances=n_max_instances, 
                is_training=self.is_training, 
            )

            eval_dict = evaluation.evaluate(
                self.pred_dict, 
                self.gt_dict, 
                is_eval=False,
                is_nn=True
            )
            self.collect_losses(eval_dict['loss_dict'])

            learning_rate = self.get_learning_rate(
                1e-3,
                self.global_step,
                self.batch_size,
                200000,
                0.7)

            update_ops = tf.get_collection(tf.GraphKeys.UPDATE_OPS)
            with tf.control_dependencies(update_ops):
                self.train_op = self.create_train_op(learning_rate, self.total_loss)

            self.saver = tf.train.Saver(max_to_keep=3)

    def create_train_op(self, learning_rate, total_loss):
        # Skip gradient update if any gradient is infinite. This should not happen and is for debug only.
        optimizer = tf.train.AdamOptimizer(learning_rate=learning_rate)
        self.optimizer = optimizer
        grads_and_vars = optimizer.compute_gradients(total_loss)
        grads = [g for g, v in grads_and_vars]
        varnames = [v for g, v in grads_and_vars]
        is_finite = tf.ones(dtype=tf.bool, shape=[])
        for g, v in grads_and_vars:
            if g is not None:
                g_is_finite = tf.reduce_any(tf.is_finite(g))
                g_is_finite_cond = tf.cond(g_is_finite, tf.no_op, lambda: tf.Print(g_is_finite, [g], '{} is not finite:'.format(str(g))))
                with tf.control_dependencies([g_is_finite_cond]):
                    is_finite = tf.logical_and(is_finite, g_is_finite)
        train_op = tf.cond(
            is_finite, 
            lambda: optimizer.apply_gradients(zip(grads, varnames), global_step=self.global_step), 
            lambda: tf.Print(is_finite, [is_finite], 'Some gradients are not finite! Skipping gradient backprop.')
        )
        return train_op

    def create_feed_dict(self, batch, is_training):
        ############################
        # TODO Task 2.2: Create the dictionary to feed the batched data
        feed_dict = {}
        for primitive in primitives.get_all_primitives_classes():
            primitive.fill_gt_placeholders(feed_dict, gt_dict['parameters'], batch)
        ############################
        return feed_dict

    def collect_losses(self, loss_dict):
        self.total_loss = tf.zeros(shape=[], dtype=tf.float32)

        self.normal_loss_per_data = loss_dict['normal_loss']
        self.total_normal_loss = tf.reduce_mean(self.normal_loss_per_data)
        normal_loss_multiplier = 1.0
        if normal_loss_multiplier > 0:
            self.total_loss += normal_loss_multiplier * self.total_normal_loss

        self.type_loss_per_data = loss_dict['type_loss']
        self.total_type_loss = tf.reduce_mean(self.type_loss_per_data)
        type_loss_multiplier = 1.0
        if type_loss_multiplier > 0:
            self.total_loss += type_loss_multiplier * self.total_type_loss

        self.miou_loss_per_data = loss_dict['avg_miou_loss']
        self.miou_loss_per_instance = loss_dict['miou_loss']
        self.total_miou_loss = tf.reduce_mean(self.miou_loss_per_data)
        miou_loss_multiplier = 1.0
        if miou_loss_multiplier > 0:
            self.total_loss += miou_loss_multiplier * self.total_miou_loss

        self.residue_loss_per_data = loss_dict['avg_residue_loss']
        self.residue_loss_per_instance = loss_dict['residue_loss']
        self.total_residue_loss = tf.reduce_mean(self.residue_loss_per_data)
        residue_loss_multiplier = 1.0
        if residue_loss_multiplier > 0:
            self.total_loss += residue_loss_multiplier * self.total_residue_loss

        self.parameter_loss_per_data = loss_dict['avg_parameter_loss']
        self.parameter_loss_per_instance = loss_dict['parameter_loss']
        self.total_parameter_loss = tf.reduce_mean(self.parameter_loss_per_data)
        parameter_loss_multiplier = 1.0
        if parameter_loss_multiplier > 0:
            self.total_loss += parameter_loss_multiplier * self.total_parameter_loss

        self.total_loss *= 1.0

    def train(self, sess, train_data, val_data, n_epochs, val_interval, snapshot_interval, model_dir, log_dir, val_pred_dir):
        assert n_epochs > 0

        if not os.path.exists(model_dir): 
            os.makedirs(model_dir)
        if not os.path.exists(val_pred_dir):
            os.makedirs(val_pred_dir)
        print('Training started.')

        start_time = time.time()
        for epoch in range(1, n_epochs + 1):
            for batch in train_data.create_iterator():
                feed_dict = self.create_feed_dict(batch, is_training=True)
                step, _, loss = sess.run([self.global_step, self.train_op, self.total_loss], feed_dict=feed_dict)

                elapsed_min = (time.time() - start_time) / 60
                print('Epoch: {:d} | Step: {:d} | Batch Loss: {:6f} | Elapsed: {:.2f}m'.format(epoch, step, loss, elapsed_min))

                if step % val_interval == 0:
                    print('Start validating...')
                    msg = 'Epoch: {:d} | Step: {:d}'.format(epoch, step)

                    remain_min = (n_epochs * train_data.n_data - step) * elapsed_min / step

                    predict_result = self.predict_and_save(sess, val_data, save_dir=os.path.join(val_pred_dir, 'step{}'.format(step)))
                    msg = predict_result['msg']
                    msg = 'Validation: ' + msg + ' | Elapsed: {:.2f}m, Remaining: {:.2f}m'.format(elapsed_min, remain_min)
                    print(msg)
                    # clean up old predictions

                if step % snapshot_interval == 0:
                    print('Saving snapshot at step {:d}...'.format(step))
                    self.saver.save(sess, os.path.join(model_dir, 'tf_model.ckpt'), global_step=step)
                    print('Done saving model at step {:d}.'.format(step))

        elapsed_min = (time.time() - start_time) / 60
        print('Training finished.')
        print('Elapsed: {:.2f}m.'.format(elapsed_min))
        print('Saved {}.'.format(self.saver.save(sess, os.path.join(model_dir, 'tf_model.ckpt'), global_step=step)))
        
    def get_learning_rate(self, init_learning_rate, global_step, batch_size, decay_step, decay_rate):
        learning_rate = tf.train.exponential_decay(
            init_learning_rate,
            global_step*batch_size,
            decay_step,
            decay_rate,
            staircase=True)
        return learning_rate

    def predict_and_save(self, sess, dset, save_dir):
        print('Predicting and saving predictions to {}...'.format(save_dir))
        losses = {
            'total_loss': 0.0, 
            'total_miou_loss': 0.0, 
            'total_normal_loss': 0.0,
            'total_type_loss': 0.0,
            'total_residue_loss': 0.0, 
            'total_parameter_loss': 0.0
        }
        if not os.path.exists(save_dir):
            os.makedirs(save_dir)
        for batch in dset.create_iterator():
            feed_dict = self.create_feed_dict(batch, is_training=False)
            loss_dict = {
                'total_loss': self.total_loss,
                'total_miou_loss': self.total_miou_loss,
                'total_normal_loss': self.total_normal_loss,
                'total_type_loss': self.total_type_loss,
                'total_residue_loss': self.total_residue_loss,
                'total_parameter_loss': self.total_parameter_loss,
            }
            pred_result, loss_result = sess.run([self.pred_dict, loss_dict], feed_dict=feed_dict)

            for key in losses.keys():
                losses[key] += loss_result[key] * dset.last_step_size
            print('Finished {}/{}'.format(dset.get_last_batch_range()[1], dset.n_data), end='\r')
        losses.update((x, y / dset.n_data) for x, y in losses.items())
        msg = self.format_loss_result(losses)
        open(os.path.join(save_dir, 'test_loss.txt'), 'w').write(msg)
        return {
            'msg': msg,
        }

    def format_loss_result(self, losses):
        msg = ''
        msg += 'Total Loss: {:6f}'.format(losses['total_loss'])
        msg += ', MIoU Loss: {:6f}'.format(losses['total_miou_loss'])
        msg += ', Normal Loss: {:6f}'.format(losses['total_normal_loss'])
        msg += ', Type Loss: {:6f}'.format(losses['total_type_loss'])
        msg += ', Parameter Loss: {:6f}'.format(losses['total_parameter_loss'])
        msg += ', Residue Loss: {:6f}'.format(losses['total_residue_loss'])
        return msg
        
